#include "DAQDataFormats/RawMilliSlice.h"
#include "DAQDataFormats/DataFormatException.h"
#include <sys/time.h>          // gettimeofday
#include <stdarg.h>            // va_start
#include <cstdio>

#include <boost/format.hpp>

namespace daqdataformats{

//------------------------------------------------------------------------------
RawMilliSlice::RawMilliSlice (const uint32_t size, const uint32_t tlvl) :
     RawDAQData(0
               ,(const init_t)            &RawMilliSlice::init
               ,(const sizeofdata_t)      &RawMilliSlice::predefinedSize
               ,(const setBufferSource_t) &RawMilliSlice::setBufferSource
               ,(const print_t)           &RawMilliSlice::printStandard
               ,(const readData_t)        &RawMilliSlice::readDataFromBuffer)
    ,_Header                  (NULL)
    ,_Index                   (NULL)
    ,_FloatingMicroSlice      (NULL)
    ,_FloatingMicroSliceNumber(-1)
    ,_fd_read_bytes           (0)
    ,_tlvl                    (tlvl)
{
    reserveInternalBuffer(size);
    defaultConstructorFloatingSize();
}

//------------------------------------------------------------------------------
RawMilliSlice::RawMilliSlice(const RawMilliSlice& copy_in) :
     RawDAQData(copy_in)
    ,_Header                  (NULL)
    ,_Index                   (NULL)
    ,_FloatingMicroSlice      (NULL)
    ,_FloatingMicroSliceNumber(-1)
    ,_fd_read_bytes           (0)
    ,_tlvl                    (copy_in._tlvl)
{
    lastOperationsInDeepCopy(copy_in);
}

//------------------------------------------------------------------------------
RawMilliSlice::RawMilliSlice(const void* millislice_buffer, const uint32_t size):
     RawDAQData(0
               ,(const init_t)            &RawMilliSlice::init
               ,(const sizeofdata_t)      &RawMilliSlice::predefinedSize
               ,(const setBufferSource_t) &RawMilliSlice::setBufferSource
               ,(const print_t)           &RawMilliSlice::printStandard
               ,(const readData_t)        &RawMilliSlice::readDataFromBuffer)
    ,_Header                  (NULL)
    ,_Index                   (NULL)
    ,_FloatingMicroSlice      (NULL)
    ,_FloatingMicroSliceNumber(-1)
    ,_fd_read_bytes           (0)
    ,_tlvl                    (0)
{

    reserveInternalBuffer(size);
    copy(millislice_buffer);
}
//------------------------------------------------------------------------------
// Destructor
 RawMilliSlice::~RawMilliSlice() {
     if(_Header)             delete _Header;
     if(_Index)              delete _Index;
     if(_FloatingMicroSlice) delete _FloatingMicroSlice;
 }

//------------------------------------------------------------------------------
// Initialization of Data members
void RawMilliSlice::init(){

    /// This is needed for the formats with variable size
    /// If buffer is internal, need to resize it
    if(isInternalBuffer())
        _InternalBuffer.resize(rawmillihdr::MILLISLICEHEADER_SIZE);

    /// Initialize the header
    _Header->init();
    
    /// Initialize the _FloatingMicroSliceNumber
    _FloatingMicroSliceNumber = -1;
}


//------------------------------------------------------------------------------
// Add a MicroSlice into the MilliSlice changing the Header appropriately
bool RawMilliSlice::addMicroSlice(const void* microslice_buffer, const bool indexate, const bool check_mc_flag){

    /// Construct a temporary nanoslice with external buffer
    const RawMicroSlice  tmp_nanoslice(microslice_buffer);
    /// This is its size
    const uint32_t microslice_size = tmp_nanoslice.sizeofdata();

    // Add one microslice
    addData(microslice_buffer, microslice_size, 1, false);

    /// Add an index if told so.
    if(indexate){
        /// Get Time length and offset
        const uint32_t time   = tmp_nanoslice.getTime() - getTimeStart();
        const uint32_t offset = sizeofdata() - rawmillihdr::MILLISLICEHEADER_SIZE;

        /// If MilliSlice Index is not constructed yet, do it now
        if(!_Index) _Index = new RawMilliSliceIndex();

        // Add the index
        _Index->addIndex(time, offset);
    }// end of adding index

    // Check if the MicroSlice is MC generated
    // If it is, set the appropriate MilliSlice MC flag
    if(check_mc_flag){ setMonteCarloFlag(tmp_nanoslice.getMonteCarloFlag());}

    return true;
}

//------------------------------------------------------------------------------
// Add a chunk of MicroSlices into the MilliSlice changing the Header appropriately
bool RawMilliSlice::addData(const void* buffer, const uint32_t size, const uint32_t number_of_microslices, const bool check_mc_flag){

    /// Just read the chunk of Microslices straight from the buffer into the end of the (internal) Buffer
    readDataGeneral(buffer, size);

    checkBufferInternalness();

    /// Advance the MilliSlice frame number (MicroSlice number in the MilliSlice)
    _Header->setMicrosliceNumber(_Header->getMicrosliceNumber() + number_of_microslices);
    /// Change the MilliSlice size
    _Header->addSliceSize(size);

    /// Check if the first MicroSlice is MC generated
    /// If it is, set the appropriate MilliSlice MC flag
    /// checking/setting MC flag making it consistent with the MicroSlice
    if(check_mc_flag){
        RawMicroSliceHeader micro_header(buffer);
        setMonteCarloFlag(micro_header.getMonteCarloFlag());
    }// end of checking/setting MC flag making it consistent with the MicroSlice

    return true;
}

//------------------------------------------------------------------------------
uint32_t RawMilliSlice::getCRC() const{

    if(!_Header->isCRCCalculationUsed()) return 0;

    return RAWBUFF32[sizeofdata() - 1];
}



//------------------------------------------------------------------------------
// Closing the MilliSlice by adding the Index (if available), tail and the CRC
bool RawMilliSlice::close(const bool produce_crc){

    PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" 0 ")
    /// First add the MilliSliceIndex if available
    if(_Index){
        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" 1 ")

        /// Close the index - produce the tail
        _Index->close();
        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" 2 ")

        const uint32_t size_of_index = _Index->sizeofdata();

        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" 3 ")
        // Copy the index to the end
        readDataGeneral(_Index->getBuffer(), size_of_index);
        checkBufferInternalness();
        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" 4 ")

        /// Adjust the MilliSlice size
        _Header->addSliceSize(size_of_index);

        /// Set MilliSlice Index number in the header
        _Header->setMilliSliceIndexNumber(_Index->getIndexNumber());

    }// end of adding MilliSliceIndex

    const uint32_t size_before_adding_tail_and_crc = sizeofdata();

    /// Remember the current Buffer position
    const void* current_buffer = &RAWBUFF32[size_before_adding_tail_and_crc];

    /// Add MILLISLICEHEADER_SIZE + 1 words to tthe size.
    /// We will add the tail (another MilliSlice header) and the CRC-word in the end of the MilliSlice
    _Header->addSliceSize(rawmillihdr::MILLISLICEHEADER_SIZE_PLUS_ONE);
    /// If CRC is calculated, set the isCRCCalculationUsed flag
    if(produce_crc)_Header->setCRCCalculationUsed(true);

    /// Adding a tail
    /// Create a new Tail and assigning its pointer
    readDataGeneral(_Header->getBuffer(), rawmillihdr::MILLISLICEHEADER_SIZE, size_before_adding_tail_and_crc);
    checkBufferInternalness();

    /// Make it a tail
    RawMilliSliceHeader tail_tmp(current_buffer);
    /// Set an appropriate Marker for the tail
    tail_tmp.setMarker(rawmillihdr::Marker_value[1]);

    // Finally, add CRC at the very end of _Buffer if requested
    if(produce_crc)
        addCheckSum();
    else{
        /// If not, we still need to make zero at the end
        /// Check if we need to assign or push_back
        if(isInternalBuffer() && _InternalBuffer.size() < sizeofdata())
            _InternalBuffer.push_back(0);
        else
            RAWBUFF32[sizeofdata() - 1] = 0;
    }

    return true;
}   // close

//------------------------------------------------------------------------------
// This is a general method to read the microslice from millislice_buffer to _Buffer
void* RawMilliSlice::readDataFromBuffer(const void* millislice_buffer)
{
    /// Read the header and find out the size of the millislice
    void* millislice_buffer_new = readDataGeneral(millislice_buffer, rawmillihdr::MILLISLICEHEADER_SIZE, 0);


    /// No need to check if the MilliSlice internal buffer went out.
    /// MilliSlice size is at least of the size of the header

    /// Now, read the rest of the millislice
    millislice_buffer_new = readDataGeneral(millislice_buffer_new, sizeofdata() - rawmillihdr::MILLISLICEHEADER_SIZE, rawmillihdr::MILLISLICEHEADER_SIZE);

    /// Check pointers after each read or resize buffer
    checkBufferInternalness();

    return millislice_buffer_new;
}// end of readData(void*)


// Method to read data in to a file descriptor
// The assumption is that the read will not block because a poll or select
// (in the case that the fd is a socket) has determined that data is ready.
// (Of course, the fd is always "ready" if the fd is associate with a file
// on disk).  But no assumption is made that a 2nd read will not block, so
// only one read (within this readData routine) is every done. The caller
// is responsible for doing the necessary number of readData calls to get a
// "complete" millislice.
readDataStatus RawMilliSlice::readData(const int fd)
{
    if (fd >= 0)
    {
        if (isInternalBuffer())
        {
	    size_t bytes_to_read;

            if (_fd_read_bytes < rawmillihdr::MILLISLICEHEADER_SIZE*sizeof(uint32_t))
            {   const size_t min_words = rawmillihdr::MILLISLICEHEADER_SIZE;
                const size_t min_bytes = min_words*sizeof (uint32_t);
                if (_InternalBuffer.size() < min_words)
                {
		    _InternalBuffer.resize(min_words);
                    resetBufferSource();
                }

		bytes_to_read = min_bytes - _fd_read_bytes;
            }
            else
            {
                uint32_t ss=sizeofdata();
		// check before using size
		if (   (getMarker() != rawmillihdr::Marker_value[0])
		    || (ss           > 0x400000/*arbitrary val*/) )
		{   tprintf(30,"no marker or size too big\n");
		    return rds_error;
		}
                if (_InternalBuffer.size() < (size_t)ss)
                {   _InternalBuffer.resize( ss );
                    resetBufferSource();
                }
                bytes_to_read = (size_t)ss*sizeof(uint32_t) - _fd_read_bytes;
            }

	    const ssize_t ss = read(fd, (char*)_Buffer+_fd_read_bytes, bytes_to_read );
	    if (ss==-1 || ss==0)
	    {   tprintf(30,"read returned %ld\n", (long int)ss);
		return rds_error;
	    }
	    _fd_read_bytes += ss;
	    if (ss != (int)bytes_to_read)
	    {   return rds_incomplete;
	    }

            const uint32_t size_of_data = sizeofdata();

	    // at this point, the header should have been read in and the
	    // size from within can be examined.
	    // check for resonableness
	    if (   (getMarker() != rawmillihdr::Marker_value[0])
		|| (size_of_data > 0x400000/*arbitrary val*/) )
	    {   tprintf(30,"no marker or size too big (2)\n");
		return rds_error;
	    }

	    if (((size_t)size_of_data*sizeof (uint32_t)) == _fd_read_bytes)
		return rds_complete;
	    return rds_incomplete;
        }
        else
        {
	    printf("external buffer\n");
            // assume there is enough space???
	    return RawDAQData::readData(fd);
        }
    }

    return rds_error;
}   // readData( int )


//------------------------------------------------------------------------------
bool RawMilliSlice::clear(){
    _fd_read_bytes = 0;

    return RawDAQData::clear();
}

//------------------------------------------------------------------------------
bool RawMilliSlice::softclear(){
    _fd_read_bytes = 0;
    return true;
}

//------------------------------------------------------------------------------
bool RawMilliSlice::setFloatingMicroSlice(const uint32_t imicro) const{

        /// Check whether MicroSlice is valid
        if(imicro >= getMicrosliceNumber())
            return false;

        /// If it's already the same, nothing needs to be done
        if((int)imicro == _FloatingMicroSliceNumber)
            return true;

        /// Pointer that we are going to use
        uint32_t* pointer;

        /// If microslice number is greater than current floating Microslice number, then no need
        /// to loop over all MicroSlices, just start from the current floating MicroSlice
        if((int)imicro > _FloatingMicroSliceNumber && _FloatingMicroSliceNumber > -1){
            pointer = (uint32_t*)_FloatingMicroSlice->getBuffer();
        }
        else{
            /// Otherwise, start from scratch
            /// First point to the first MicroSlice
            pointer = (uint32_t*)&RAWBUFF32[rawmillihdr::MILLISLICEHEADER_SIZE];

            /// Set floating MicroSlice to pointer
            /// If it doesn't exist, create it
            setFloatingMicroSliceToBufferSource(pointer);
            _FloatingMicroSliceNumber = 0;
        }// end of setting the first floating MicroSlice

        /// Pointer to the end of the buffer for checks
        const uint32_t* end_of_buffer_pointer = &RAWBUFF32[sizeofdata()];

        /// Start loop over microslices
        for(; _FloatingMicroSliceNumber < (int)imicro; ++_FloatingMicroSliceNumber){
            pointer += _FloatingMicroSlice->sizeofdata();

            /// Check if the pointer is not valid
            if (pointer >= end_of_buffer_pointer)
                return false;

            _FloatingMicroSlice->setBufferSource(pointer);
        }

        return true;
}

//------------------------------------------------------------------------------
bool RawMilliSlice::setFloatingMicroSliceToBufferSource(const void* buffer) const{

            /// If it exists, just set its buffer
            if(_FloatingMicroSlice)
                return _FloatingMicroSlice->setBufferSource(buffer);

            /// Otherwise, create new Floating MicroSlice pointing to the buffer
            _FloatingMicroSlice = new RawMicroSlice(buffer);

            return true;
}

//------------------------------------------------------------------------------
// Printing the MilliSlice structure
void RawMilliSlice::printStandard(std::ostream& os) const{

    _Header->print(os);

    const uint32_t size_of_data = sizeofdata();

    os<<boost::format("\n---MilliSlice---------------------------------------------size = %u words\n")% size_of_data;

    os<<"\nMilliSlice Data RAW 32bit WORDs (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        this->printWord(i, false, os);
        os<<" ";
    }

    os<<"\nMilliSlice Data RAW 32bit Bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
            os<<"\n    ";
            this->printWord(i, true, os);
    }

    const uint32_t num_microslices = getMicrosliceNumber();
    os<<std::endl<<"----------NumMicroSlices-------="<<num_microslices<<std::endl;

    /// If number is nonzero, need to print them
    if(num_microslices){
        /// Get a copy of the MilliSlice this is so that we don't screw up the internal structure
        RawMilliSlice milli_copy(_Buffer, 0);

        for(uint32_t i = 0; i < num_microslices; ++i){
            milli_copy.setFloatingMicroSlice(i);
            milli_copy.getFloatingMicroSlice()->print(os);
        }// end of loop over nanoslices
    }// end of checking whether the number of nanoslices is non-zero

}

//------------------------------------------------------------------------------
bool RawMilliSlice::setBufferSource(const void* pointer){

   _Buffer = (void*)pointer;

   _shouldBufferBeInternal = isInternalBuffer();

   /// If Header is already defined, then set its buffer source
   /// Otherwise, need to construct it
   if(_Header)return _Header->setBufferSource(pointer);

   _Header = new RawMilliSliceHeader(_Buffer); ///< version of the MilliSLiceSliceHeader equals the version of the MilliSlice

   _function_sizeofdata = (sizeofdata_t) &RawMilliSlice::sizeofdata;

   return true;
}

//------------------------------------------------------------------------------
void RawMilliSlice::tprintf( unsigned lvl, const char *fmt, ... ) {

    if (lvl > 31) lvl=31;
    if (_tlvl & (1<<lvl))
    {   timeval 	tv;
	char            str[80];
	time_t          local;
	const char*     indent="                                ";//32 spaces
	va_list         ap;
	va_start( ap, fmt);
	gettimeofday( &tv, NULL );
	local = tv.tv_sec;
	strftime( str, sizeof(str), "%T", localtime(&local) );
	printf( "%s.%06ld%s", str, (long)tv.tv_usec, &indent[31-lvl] );
	vprintf( fmt, ap );
    }
}   // tprintf

//------------------------------------------------------------------------------
daqdataformats::RawMilliSliceIndex*  RawMilliSlice::getMilliSliceIndex() const {

    /// If there are no MilliSlices Indeces in the header,
    /// then nothing to return - MilliSlice Index was not constructed
    if(getMilliSliceIndexNumber() == 0)
        return NULL;

    /// If the MilliSlice Index does not exist,
    /// then we need to figure out its pointer and construct the Index
    if(!_Index){
            /// At this point it is assumed that the MilliSlice is fully closed

            /// Get the tail. It's the end of the buffer, minus CRC and MilliSliceTail
            /// Actually it's only tail, but we declare it as the MilliSliceIndex, because we need to know the size of the whole MilliSliceIndex
            const uint32_t*          pointer_to_tail = &RAWBUFF32[sizeofdata() - rawmillihdr::MILLISLICEHEADER_SIZE_PLUS_ONE];
            const RawMilliSliceIndex tail(pointer_to_tail);

            /// The true pointer to Index is pointer to the tail minus the size
            const uint32_t* pointer_to_millislice_index = pointer_to_tail + 1 - tail.sizeofdata();

            /// Construct the Index
            _Index = new RawMilliSliceIndex(pointer_to_millislice_index);
    }// end of checking whether Index is not constructed

    return _Index;
}

} // end of namespace
