#include <DatabaseUtils/DAQConfig/DDTConnectConfig.h>
#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDatabase/Table.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/algorithm/string.hpp>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

/**
 * Creates a DDT Connection configuration object from the specified
 * list of subsystem configuration IDs.
 *
 * @param subsystemConfigIdList The list of subsystem
 *        configuration IDs.
 */
  DDTConnectConfig::
  DDTConnectConfig(const std::vector<int64_t>& subsystemConfigIdList)
  {
    _primarySubsystemConfigId = -1;
    if (subsystemConfigIdList.size() > 0) {
      _primarySubsystemConfigId = subsystemConfigIdList[0];
      _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
      _hasValidData = false;
    }
  }

  /**
   * Creates a DDT Connection configuration object from the specified
   * subsystem configuration specification.
   *
   * @param nameSpec The named configuration of interest.
   */
  DDTConnectConfig::DDTConnectConfig(const SubsystemNameSpec& nameSpec)
  {
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList)) {
      _primarySubsystemConfigId = subsystemConfigIdList[0];
      _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
      _hasValidData = false;
    }
  }

  /**
   * Creates a DDT Connection configuration object from the specified
   * global configuration specification.
   *
   * @param nameSpec The named configuration of interest.
   */
  DDTConnectConfig::DDTConnectConfig(const GlobalNameSpec& nameSpec)
  {
    _primarySubsystemConfigId = -1;
    int64_t subsystemConfigId;
    bool status = NamedConfigUtils::
      getSubsystemConfigIdFromGlobalName(nameSpec,
					 SubsystemNameSpec::
					 DDT_SUBSYSTEM_NAME,
					 SubsystemNameSpec::
					 CONNECT_CONFIG_TYPE,
					 subsystemConfigId);

    std::vector<int64_t> subsystemConfigIdList;
    if (status) {
      IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
      status = NamedConfigUtils::
	getConfigIdAncestry(idSpec, subsystemConfigIdList);
    }

    if (status) {
      _primarySubsystemConfigId = subsystemConfigIdList[0];
      _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
      _hasValidData = false;
    }
  }

  /**
   * Creates a DDT Connection configuration object from the specified
   * global configuration ID.
   *
   * @param globalConfigId The ID of the global configuration of interest.
   */
  DDTConnectConfig::DDTConnectConfig(const int64_t& globalConfigId)
  {
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 DDT_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 CONNECT_CONFIG_TYPE,
                                                 subsystemConfigIdList)) {
      _primarySubsystemConfigId = subsystemConfigIdList[0];
      _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
      _hasValidData = false;
    }
  }

  bool DDTConnectConfig::
  _fetchAllData(const std::vector<int64_t>& subsystemConfigIdList)
  {
    std::string table_name = 
      "NovaDatabase/tables/DAQConfig/DDTManagerConnectParameters.xml";

    ConfigDataSelector cfgSelector;
    cfgSelector.addSelectionColumn(DBColumnParams::HOSTNAME_COLUMN_NAME);

    std::vector<std::string> dataColumnsOfInterest;
    dataColumnsOfInterest.push_back("configPath");
    dataColumnsOfInterest.push_back("nProcesses");
    dataColumnsOfInterest.push_back("version");
    dataColumnsOfInterest.push_back("qualifier");

    _ddtConfigDataTree.reset
      (new ConfigDataTree(table_name, subsystemConfigIdList,
			  cfgSelector, dataColumnsOfInterest));

    return _ddtConfigDataTree->hasValidData();
  }
};  // end of namespace daqconfig
};  // end of namespace dbutils
