#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <boost/lexical_cast.hpp>

namespace dbutils {
namespace daqconfig {

/**
 * Creates a global name specification object from the specified
 * configuration name DCM operating mode, and detector ID.
 *
 * @param configName The named global configuration of interest.
 * @param dcmOperatingMode The operating mode of interest (e.g. DCS or DSO).
 * @param detectorId The detector of interest (e.g. NDOS or FarDet).
 */
GlobalNameSpec::
GlobalNameSpec(const std::string& configName,
               const std::string& dcmOperatingMode,
               const std::string& detectorId) :
    _configName(configName), _dcmOperatingMode(dcmOperatingMode),
    _detectorId(detectorId), _hasTimestamp(false),
    _skipInvalidStateTest(false), _isSlowControlsConfig(false),
    _usageCount(0), _sortIndex(-1)
{
}

/**
 * Creates a global name specification object from the specified
 * configuration name, DCM operating mode, detector ID, and timestamp.
 *
 * @param configName The named global configuration of interest.
 * @param dcmOperatingMode The operating mode of interest (e.g. DCS or DSO).
 * @param detectorId The detector of interest (e.g. NDOS or FarDet).
 * @param timestamp The date/time string that should be used to
 *        determine the time-based validity context for the named
 *        configuration.  Supported formats include "yyyy-mm-dd HH:MM:SS".
 */
GlobalNameSpec::
GlobalNameSpec(const std::string& configName,
               const std::string& dcmOperatingMode,
               const std::string& detectorId,
               const std::string& timestamp) :
    _configName(configName), _dcmOperatingMode(dcmOperatingMode),
    _detectorId(detectorId), _hasTimestamp(true), _timestamp(timestamp),
    _skipInvalidStateTest(false), _isSlowControlsConfig(false),
    _usageCount(0), _sortIndex(-1)
{
}

/**
 * Creates a string representation of the name specification.
 */
std::string GlobalNameSpec::toString(int overallIndent) const
{
    std::string workString;
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("Named Global Configuration:\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  configuration name: " + _configName + "\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  DCM operating mode: " + _dcmOperatingMode + "\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  detector name: " + _detectorId + "\n");

    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    if (_hasTimestamp) {
        workString.append("  timestamp: " + _timestamp);
    }
    else {
        workString.append("  no timestamp");
    }
    workString.append("\n");

    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  usage count: " +
                      boost::lexical_cast<std::string>(_usageCount) + "\n");

    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  DAQ or Slow Controls: ");
    workString.append((_isSlowControlsConfig ? "Slow Controls" : "DAQ"));

    if (_skipInvalidStateTest) {
        workString.append("\n");
        if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
        workString.append("  Skipping the Invalid State tests");
    }

    return workString;
}

/**
 * Tests if this instance is less than the specified one.
 */
bool GlobalNameSpec::
operator<(const GlobalNameSpec& other) const
{
    if (this->_sortIndex > -1 && other._sortIndex > -1) {
        return (this->_sortIndex < other._sortIndex);
    }

    if (this->_usageCount > other._usageCount) {return false;}
    if (this->_usageCount < other._usageCount) {return true;}
    // else _usageCount is equal

    if (this->_detectorId > other._detectorId) {return false;}
    if (this->_detectorId < other._detectorId) {return true;}
    // else _detectorId is equal

    if (this->_configName > other._configName) {return false;}
    if (this->_configName < other._configName) {return true;}
    // else _configName is equal

    if (this->_dcmOperatingMode > other._dcmOperatingMode) {return false;}
    if (this->_dcmOperatingMode < other._dcmOperatingMode) {return true;}
    // else _dcmOperatingMode is equal

    if (this->_timestamp > other._timestamp) {return false;}
    if (this->_timestamp < other._timestamp) {return true;}
    // else _timestamp is equal

    return false;
}

/**
 * Adds a specified number of spaces to a string.
 */
void GlobalNameSpec::_addSpaces(std::string& workString, int count) const
{
    for (int idx = 0; idx < count; ++idx) {
        workString.append(" ");
    }
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
